from panther.db.models import Model


def get_model_fields(model):
    result = {}

    for k, v in model.model_fields.items():
        try:
            is_model = issubclass(v.annotation, Model)
        except TypeError:
            is_model = False

        if is_model:
            result[k] = get_model_fields(v.annotation)
        else:
            result[k] = getattr(v.annotation, '__name__', str(v.annotation))
    return result



import asyncio
import contextlib
import logging
import sys
import types
from collections.abc import Callable
from logging.config import dictConfig
from pathlib import Path

import orjson as json

import panther.logging
from panther import status
from panther._load_configs import *
from panther._utils import clean_traceback_message, reformat_code, check_class_type_endpoint, check_function_type_endpoint
from panther.cli.utils import print_info
from panther.configs import config
from panther.events import Event
from panther.exceptions import APIError, PantherError
from panther.monitoring import Monitoring
from panther.request import Request
from panther.response import Response
from panther.routings import find_endpoint

dictConfig(panther.logging.LOGGING)
logger = logging.getLogger('panther')


class Panther:
    def __init__(self, name: str, configs: str | None = None, urls: dict | None = None):
        self._configs_module_name = configs
        self._urls = urls

        config.BASE_DIR = Path(name).resolve().parent

        try:
            self.load_configs()
            if config.AUTO_REFORMAT:
                reformat_code(base_dir=config.BASE_DIR)
        except Exception as e:  # noqa: BLE001
            logger.error(e.args[0] if isinstance(e, PantherError) else clean_traceback_message(e))
            sys.exit()

        # Print Info
        print_info(config)

    def load_configs(self) -> None:
        # Check & Read The Configs File
        self._configs_module = load_configs_module(self._configs_module_name)

        load_redis(self._configs_module)
        load_startup(self._configs_module)
        load_shutdown(self._configs_module)
        load_timezone(self._configs_module)
        load_database(self._configs_module)
        load_secret_key(self._configs_module)
        load_monitoring(self._configs_module)
        load_throttling(self._configs_module)
        load_user_model(self._configs_module)
        load_log_queries(self._configs_module)
        load_middlewares(self._configs_module)
        load_auto_reformat(self._configs_module)
        load_background_tasks(self._configs_module)
        load_default_cache_exp(self._configs_module)
        load_authentication_class(self._configs_module)
        load_urls(self._configs_module, urls=self._urls)
        load_websocket_connections()

    async def __call__(self, scope: dict, receive: Callable, send: Callable) -> None:
        if scope['type'] == 'lifespan':
            message = await receive()
            if message["type"] == 'lifespan.startup':
                if config.HAS_WS:
                    await config.WEBSOCKET_CONNECTIONS.start()
                await Event.run_startups()
            elif message["type"] == 'lifespan.shutdown':
                # It's not happening :\, so handle the shutdowns in __del__ ...
                pass
            return

        func = self.handle_http if scope['type'] == 'http' else self.handle_ws
        await func(scope=scope, receive=receive, send=send)

    async def handle_ws(self, scope: dict, receive: Callable, send: Callable) -> None:
        from panther.websocket import GenericWebsocket, Websocket

        # Monitoring
        monitoring = Monitoring(is_ws=True)

        # Create Temp Connection
        temp_connection = Websocket(scope=scope, receive=receive, send=send)
        await monitoring.before(request=temp_connection)
        temp_connection._monitoring = monitoring

        # Find Endpoint
        endpoint, found_path = find_endpoint(path=temp_connection.path)
        if endpoint is None:
            logger.debug(f'Path `{temp_connection.path}` not found')
            return await temp_connection.close()

        # Check Endpoint Type
        if not issubclass(endpoint, GenericWebsocket):
            logger.critical(f'You may have forgotten to inherit from `GenericWebsocket` on the `{endpoint.__name__}()`')
            return await temp_connection.close()

        # Create The Connection
        del temp_connection
        connection = endpoint(scope=scope, receive=receive, send=send)
        connection._monitoring = monitoring

        # Collect Path Variables
        connection.collect_path_variables(found_path=found_path)

        middlewares = [middleware(**data) for middleware, data in config.WS_MIDDLEWARES]

        # Call Middlewares .before()
        await self._run_ws_middlewares_before_listen(connection=connection, middlewares=middlewares)

        # Listen The Connection
        await config.WEBSOCKET_CONNECTIONS.listen(connection=connection)

        # Call Middlewares .after()
        middlewares.reverse()
        await self._run_ws_middlewares_after_listen(connection=connection, middlewares=middlewares)

    @classmethod
    async def _run_ws_middlewares_before_listen(cls, *, connection, middlewares):
        try:
            for middleware in middlewares:
                new_connection = await middleware.before(request=connection)
                if new_connection is None:
                    logger.critical(
                        f'Make sure to return the `request` at the end of `{middleware.__class__.__name__}.before()`')
                    await connection.close()
                connection = new_connection
        except APIError as e:
            connection.log(e.detail)
            await connection.close()

    @classmethod
    async def _run_ws_middlewares_after_listen(cls, *, connection, middlewares):
        for middleware in middlewares:
            with contextlib.suppress(APIError):
                connection = await middleware.after(response=connection)
                if connection is None:
                    logger.critical(
                        f'Make sure to return the `response` at the end of `{middleware.__class__.__name__}.after()`')
                    break

    async def handle_http(self, scope: dict, receive: Callable, send: Callable) -> None:
        # Monitoring
        monitoring = Monitoring()

        request = Request(scope=scope, receive=receive, send=send)

        await monitoring.before(request=request)

        # Read Request Payload
        await request.read_body()

        # Find Endpoint
        _endpoint, found_path = find_endpoint(path=request.path)
        if _endpoint is None:
            return await self._raise(send, monitoring=monitoring, status_code=status.HTTP_404_NOT_FOUND)

        # Check Endpoint Type
        try:
            if isinstance(_endpoint, types.FunctionType):
                endpoint = check_function_type_endpoint(endpoint=_endpoint)
            else:
                endpoint = check_class_type_endpoint(endpoint=_endpoint)
        except TypeError:
            return await self._raise(send, monitoring=monitoring, status_code=status.HTTP_501_NOT_IMPLEMENTED)

        # Collect Path Variables
        request.collect_path_variables(found_path=found_path)

        middlewares = [middleware(**data) for middleware, data in config.HTTP_MIDDLEWARES]
        try:  # They Both(middleware.before() & _endpoint()) Have The Same Exception (APIError)
            # Call Middlewares .before()
            for middleware in middlewares:
                request = await middleware.before(request=request)
                if request is None:
                    logger.critical(
                        f'Make sure to return the `request` at the end of `{middleware.__class__.__name__}.before()`')
                    return await self._raise(send, monitoring=monitoring)

            # Call Endpoint
            response = await endpoint(request=request)

        except APIError as e:
            response = self._handle_exceptions(e)

        except Exception as e:  # noqa: BLE001
            # All unhandled exceptions are caught here
            exception = clean_traceback_message(exception=e)
            logger.critical(exception)
            return await self._raise(send, monitoring=monitoring)

        # Call Middlewares .after()
        middlewares.reverse()
        for middleware in middlewares:
            try:
                response = await middleware.after(response=response)
                if response is None:
                    logger.critical(
                        f'Make sure to return the `response` at the end of `{middleware.__class__.__name__}.after()`')
                    return await self._raise(send, monitoring=monitoring)
            except APIError as e:  # noqa: PERF203
                response = self._handle_exceptions(e)

        await response.send(send, receive, monitoring=monitoring)

    def __del__(self):
        Event.run_shutdowns()

    @classmethod
    def _handle_exceptions(cls, e: APIError, /) -> Response:
        return Response(
            data=e.detail if isinstance(e.detail, dict) else {'detail': e.detail},
            status_code=e.status_code,
        )

    @classmethod
    async def _raise(cls, send, *, monitoring, status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR):
        headers = [[b'Content-Type', b'application/json']]
        body = json.dumps({'detail': status.status_text[status_code]})
        await monitoring.after(status_code)
        await send({'type': 'http.response.start', 'status': status_code, 'headers': headers})
        await send({'type': 'http.response.body', 'body': body, 'more_body': False})




from collections import namedtuple
from datetime import timedelta, datetime
import logging
from types import NoneType

import orjson as json

from panther.configs import config
from panther.db.connections import redis
from panther.request import Request
from panther.response import Response, ResponseDataTypes
from panther.throttling import throttling_storage
from panther.utils import generate_hash_value_from_string, round_datetime

logger = logging.getLogger('panther')

caches = {}
CachedResponse = namedtuple('CachedResponse', ['data', 'status_code'])


def api_cache_key(request: Request, cache_exp_time: timedelta | None = None) -> str:
    client = request.user and request.user.id or request.client.ip
    query_params_hash = generate_hash_value_from_string(request.scope['query_string'].decode('utf-8'))
    key = f'{client}-{request.path}-{query_params_hash}-{request.validated_data}'

    if cache_exp_time:
        time = round_datetime(datetime.now(), cache_exp_time)
        return f'{time}-{key}'

    return key


def throttling_cache_key(request: Request, duration: timedelta) -> str:
    client = request.user and request.user.id or request.client.ip
    time = round_datetime(datetime.now(), duration)
    return f'{time}-{client}-{request.path}'


async def get_response_from_cache(*, request: Request, cache_exp_time: timedelta) -> CachedResponse | None:
    """
    If redis.is_connected:
        Get Cached Data From Redis
    else:
        Get Cached Data From Memory
    """
    if redis.is_connected:
        key = api_cache_key(request=request)
        data = (await redis.get(key) or b'{}').decode()
        if cached_value := json.loads(data):
            return CachedResponse(*cached_value)

    else:
        key = api_cache_key(request=request, cache_exp_time=cache_exp_time)
        if cached_value := caches.get(key):
            return CachedResponse(*cached_value)


async def set_response_in_cache(*, request: Request, response: Response, cache_exp_time: timedelta | int) -> None:
    """
    If redis.is_connected:
        Cache The Data In Redis
    else:
        Cache The Data In Memory
    """

    cache_data: tuple[ResponseDataTypes, int] = (response.data, response.status_code)

    if redis.is_connected:
        key = api_cache_key(request=request)

        cache_exp_time = cache_exp_time or config.DEFAULT_CACHE_EXP
        cache_data: bytes = json.dumps(cache_data)

        if not isinstance(cache_exp_time, timedelta | int | NoneType):
            msg = '"cache_exp_time" should be instance of "datetime.timedelta" or "int" or "None"'
            raise TypeError(msg)

        if cache_exp_time is None:
            logger.warning(
                'your response are going to cache in redis forever '
                '** set DEFAULT_CACHE_EXP in `configs` or set the `cache_exp_time` in `@API.get()` to prevent this **'
            )
            await redis.set(key, cache_data)
        else:
            await redis.set(key, cache_data, ex=cache_exp_time)

    else:
        key = api_cache_key(request=request, cache_exp_time=cache_exp_time)
        caches[key] = cache_data

        if cache_exp_time:
            logger.info('`cache_exp_time` is not very accurate when `redis` is not connected.')


async def get_throttling_from_cache(request: Request, duration: timedelta) -> int:
    """
    If redis.is_connected:
        Get Cached Data From Redis
    else:
        Get Cached Data From Memory
    """
    key = throttling_cache_key(request=request, duration=duration)

    if redis.is_connected:
        data = (await redis.get(key) or b'0').decode()
        return json.loads(data)

    else:
        return throttling_storage[key]


async def increment_throttling_in_cache(request: Request, duration: timedelta) -> None:
    """
    If redis.is_connected:
        Increment The Data In Redis
    else:
        Increment The Data In Memory
    """
    key = throttling_cache_key(request=request, duration=duration)

    if redis.is_connected:
        await redis.incrby(key, amount=1)

    else:
        throttling_storage[key] += 1



import typing
from typing import Any

from pydantic import create_model, BaseModel, ConfigDict
from pydantic.fields import FieldInfo, Field
from pydantic_core._pydantic_core import PydanticUndefined

from panther.db import Model
from panther.request import Request


class MetaModelSerializer:
    KNOWN_CONFIGS = ['model', 'fields', 'exclude', 'required_fields', 'optional_fields']

    def __new__(
            cls,
            cls_name: str,
            bases: tuple[type[typing.Any], ...],
            namespace: dict[str, typing.Any],
            **kwargs
    ):
        if cls_name == 'ModelSerializer':
            # Put `model` and `request` to the main class with `create_model()`
            namespace['__annotations__'].pop('model')
            namespace['__annotations__'].pop('request')
            cls.model_serializer = type(cls_name, (), namespace)
            return super().__new__(cls)

        # 1. Initial Check
        cls.check_config(cls_name=cls_name, namespace=namespace)
        config = namespace.pop('Config')

        # 2. Collect `Fields`
        field_definitions = cls.collect_fields(config=config, namespace=namespace)

        # 3. Collect `pydantic.model_config`
        model_config = cls.collect_model_config(config=config, namespace=namespace)
        namespace |= {'model_config': model_config}

        # 4. Create a serializer
        return create_model(
            __model_name=cls_name,
            __module__=namespace['__module__'],
            __validators__=namespace,
            __base__=(cls.model_serializer, BaseModel),
            model=(typing.ClassVar[type[BaseModel]], config.model),
            request=(Request, Field(None, exclude=True)),
            **field_definitions
        )

    @classmethod
    def check_config(cls, cls_name: str, namespace: dict) -> None:
        module = namespace['__module__']
        address = f'{module}.{cls_name}'

        # Check `Config`
        if (config := namespace.get('Config')) is None:
            msg = f'`class Config` is required in {address}.'
            raise AttributeError(msg) from None

        # Check `model`
        if (model := getattr(config, 'model', None)) is None:
            msg = f'`{cls_name}.Config.model` is required.'
            raise AttributeError(msg) from None

        # Check `model` type
        try:
            if not issubclass(model, (Model, BaseModel)):
                msg = f'`{cls_name}.Config.model` is not subclass of `panther.db.Model` or `pydantic.BaseModel`.'
                raise AttributeError(msg) from None
        except TypeError:
            msg = f'`{cls_name}.Config.model` is not subclass of `panther.db.Model`.'
            raise AttributeError(msg) from None

        # Check `fields`
        if not hasattr(config, 'fields'):
            msg = f'`{cls_name}.Config.fields` is required.'
            raise AttributeError(msg) from None

        if config.fields != '*':
            for field_name in config.fields:
                if field_name == '*':
                    msg = f"`{cls_name}.Config.fields.{field_name}` is not valid. Did you mean `fields = '*'`"
                    raise AttributeError(msg) from None

                if field_name not in model.model_fields:
                    msg = f'`{cls_name}.Config.fields.{field_name}` is not in `{model.__name__}.model_fields`'
                    raise AttributeError(msg) from None

        # Check `required_fields`
        if not hasattr(config, 'required_fields'):
            config.required_fields = []

        if config.required_fields != '*':
            for required in config.required_fields:
                if required not in config.fields:
                    msg = f'`{cls_name}.Config.required_fields.{required}` should be in `Config.fields` too.'
                    raise AttributeError(msg) from None

        # Check `optional_fields`
        if not hasattr(config, 'optional_fields'):
            config.optional_fields = []

        if config.optional_fields != '*':
            for optional in config.optional_fields:
                if optional not in config.fields:
                    msg = f'`{cls_name}.Config.optional_fields.{optional}` should be in `Config.fields` too.'
                    raise AttributeError(msg) from None

        # Check `required_fields` and `optional_fields` together
        if (
                (config.optional_fields == '*' and config.required_fields != []) or
                (config.required_fields == '*' and config.optional_fields != [])
        ):
            msg = (
                f"`{cls_name}.Config.optional_fields` and "
                f"`{cls_name}.Config.required_fields` can't include same fields at the same time"
            )
            raise AttributeError(msg) from None
        for optional in config.optional_fields:
            for required in config.required_fields:
                if optional == required:
                    msg = (
                        f"`{optional}` can't be in `{cls_name}.Config.optional_fields` and "
                        f"`{cls_name}.Config.required_fields` at the same time"
                    )
                    raise AttributeError(msg) from None

        # Check `exclude`
        if not hasattr(config, 'exclude'):
            config.exclude = []

        for field_name in config.exclude:
            if field_name not in model.model_fields:
                msg = f'`{cls_name}.Config.exclude.{field_name}` is not valid.'
                raise AttributeError(msg) from None

            if config.fields != '*' and field_name not in config.fields:
                msg = f'`{cls_name}.Config.exclude.{field_name}` is not defined in `Config.fields`.'
                raise AttributeError(msg) from None

    @classmethod
    def collect_fields(cls, config: typing.Callable, namespace: dict) -> dict:
        field_definitions = {}

        # Define `fields`
        if config.fields == '*':
            for field_name, field in config.model.model_fields.items():
                field_definitions[field_name] = (field.annotation, field)
        else:
            for field_name in config.fields:
                field_definitions[field_name] = (
                    config.model.model_fields[field_name].annotation,
                    config.model.model_fields[field_name]
                )

        # Apply `exclude`
        for field_name in config.exclude:
            del field_definitions[field_name]

        # Apply `required_fields`
        if config.required_fields == '*':
            for value in field_definitions.values():
                value[1].default = PydanticUndefined
        else:
            for field_name in config.required_fields:
                field_definitions[field_name][1].default = PydanticUndefined

        # Apply `optional_fields`
        if config.optional_fields == '*':
            for value in field_definitions.values():
                value[1].default = value[0]()
        else:
            for field_name in config.optional_fields:
                field_definitions[field_name][1].default = field_definitions[field_name][0]()

        # Collect and Override `Class Fields`
        for key, value in namespace.pop('__annotations__', {}).items():
            field_info = namespace.pop(key, FieldInfo(annotation=value))
            field_definitions[key] = (value, field_info)

        return field_definitions

    @classmethod
    def collect_model_config(cls, config: typing.Callable, namespace: dict) -> dict:
        return {
            attr: getattr(config, attr) for attr in dir(config)
            if not attr.startswith('__') and attr not in cls.KNOWN_CONFIGS
        } | namespace.pop('model_config', {}) | {'arbitrary_types_allowed': True}


class ModelSerializer(metaclass=MetaModelSerializer):
    """
    Doc:
        https://pantherpy.github.io/serializer/#style-2-model-serializer
    Example:
        class PersonSerializer(ModelSerializer):
            class Meta:
                model = Person
                fields = '*'
                exclude = ['created_date']  # Optional
                required_fields = ['first_name', 'last_name']  # Optional
                optional_fields = ['age']  # Optional
    """
    model: type[BaseModel]
    request: Request

    async def create(self, validated_data: dict):
        """
        validated_data = ModelSerializer.model_dump()
        """
        return await self.model.insert_one(validated_data)

    async def update(self, instance: Model, validated_data: dict):
        """
        instance = UpdateAPI.object()
        validated_data = ModelSerializer.model_dump()
        """
        await instance.update(validated_data)
        return instance

    async def partial_update(self, instance: Model, validated_data: dict):
        """
        instance = UpdateAPI.object()
        validated_data = ModelSerializer.model_dump(exclude_none=True)
        """
        await instance.update(validated_data)
        return instance

    async def prepare_response(self, instance: Any, data: dict) -> dict:
        return data



import copy
import typing
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import Callable

from pydantic._internal._model_construction import ModelMetaclass

from panther.throttling import Throttling


class JWTConfig:
    def __init__(
            self,
            key: str,
            algorithm: str = 'HS256',
            life_time: timedelta | int = timedelta(days=1),
            refresh_life_time: timedelta | int | None = None,
    ):
        self.key = key
        self.algorithm = algorithm
        self.life_time = life_time.total_seconds() if isinstance(life_time, timedelta) else life_time

        if refresh_life_time:
            if isinstance(refresh_life_time, timedelta):
                self.refresh_life_time = refresh_life_time.total_seconds()
            else:
                self.refresh_life_time = refresh_life_time
        else:
            self.refresh_life_time = self.life_time * 2


class QueryObservable:
    observers = []

    @classmethod
    def observe(cls, observer):
        cls.observers.append(observer)

    @classmethod
    def update(cls):
        for observer in cls.observers:
            observer._reload_bases(parent=config.QUERY_ENGINE)


@dataclass
class Config:
    BASE_DIR: Path
    MONITORING: bool
    LOG_QUERIES: bool
    DEFAULT_CACHE_EXP: timedelta | None
    THROTTLING: Throttling | None
    SECRET_KEY: bytes | None
    HTTP_MIDDLEWARES: list[tuple]
    WS_MIDDLEWARES: list[tuple]
    USER_MODEL: ModelMetaclass | None
    AUTHENTICATION: ModelMetaclass | None
    WS_AUTHENTICATION: ModelMetaclass | None
    JWT_CONFIG: JWTConfig | None
    MODELS: list[dict]
    FLAT_URLS: dict
    URLS: dict
    WEBSOCKET_CONNECTIONS: typing.Callable | None
    BACKGROUND_TASKS: bool
    HAS_WS: bool
    STARTUPS: list[Callable]
    SHUTDOWNS: list[Callable]
    TIMEZONE: str
    AUTO_REFORMAT: bool
    QUERY_ENGINE: typing.Callable | None
    DATABASE: typing.Callable | None

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == 'QUERY_ENGINE' and value:
            QueryObservable.update()

    def __setitem__(self, key, value):
        setattr(self, key.upper(), value)

    def __getitem__(self, item):
        return getattr(self, item.upper())

    def refresh(self):
        # In some tests we need to `refresh` the `config` values
        for key, value in copy.deepcopy(default_configs).items():
            setattr(self, key, value)


default_configs = {
    'BASE_DIR': Path(),
    'MONITORING': False,
    'LOG_QUERIES': False,
    'DEFAULT_CACHE_EXP': None,
    'THROTTLING': None,
    'SECRET_KEY': None,
    'HTTP_MIDDLEWARES': [],
    'WS_MIDDLEWARES': [],
    'USER_MODEL': None,
    'AUTHENTICATION': None,
    'WS_AUTHENTICATION': None,
    'JWT_CONFIG': None,
    'MODELS': [],
    'FLAT_URLS': {},
    'URLS': {},
    'WEBSOCKET_CONNECTIONS': None,
    'BACKGROUND_TASKS': False,
    'HAS_WS': False,
    'STARTUPS': [],
    'SHUTDOWNS': [],
    'TIMEZONE': 'UTC',
    'AUTO_REFORMAT': False,
    'QUERY_ENGINE': None,
    'DATABASE': None,
}

config = Config(**copy.deepcopy(default_configs))



import logging
import platform

from rich import print as rprint

from panther.configs import Config
from panther.exceptions import PantherError

logger = logging.getLogger('panther')

if platform.system() == 'Windows':
    h = '|'
    v = '_'
    tr = ' '
    tl = ' '
    br = ' '
    bl = ' '
else:
    h = '│'
    v = '─'
    tr = '╮'
    tl = '╭'
    br = '╯'
    bl = '╰'

top = f'{tl}{58 * v}{tr}'
bottom = f'{bl}{58 * v}{br}'

logo = rf"""{top}
{h}    ____                 __    __                         {h}
{h}   /\  _`\              /\ \__/\ \                        {h}
{h}   \ \ \L\ \ __      ___\ \ ,_\ \ \___      __   _ __     {h}
{h}    \ \ ,__/'__`\  /' _ `\ \ \/\ \  _ `\  /'__`\/\`'__\   {h}
{h}     \ \ \/\ \L\.\_/\ \/\ \ \ \_\ \ \ \ \/\  __/\ \ \/    {h}
{h}      \ \_\ \__/.\_\ \_\ \_\ \__\\ \_\ \_\ \____\\ \_\    {h}
{h}       \/_/\/__/\/_/\/_/\/_/\/__/ \/_/\/_/\/____/ \/_/    {h}
{h}                                                          {h}"""

help_message = f"""{logo}
{h}   usage:                                                 {h}
{h}       - panther create                                   {h}
{h}           Create project interactive                     {h}
{h}                                                          {h}
{h}       - panther create <project_name> <directory>        {h}
{h}           Default<directory> is `.`                      {h}
{h}           * It will create the project non-interactive   {h}
{h}                                                          {h}
{h}       - panther run [--reload | --help]                  {h}
{h}           Run your project with uvicorn                  {h}
{h}                                                          {h}
{h}       - panther shell [ bpython | ipython ]              {h}
{h}           Run interactive python shell                   {h}
{h}                                                          {h}
{h}       - panther monitor                                  {h}
{h}           Show the monitor :)                            {h}
{h}                                                          {h}
{h}       - panther version                                  {h}
{h}           Print the current version of Panther           {h}
{h}                                                          {h}
{h}       - panther h | help | --help | -h                   {h}
{h}           Show this message and exit                     {h}
{bottom}
"""


def import_error(message: str | Exception, package: str | None = None) -> PantherError:
    msg = str(message)
    if package:
        msg += f' -> Hint: `pip install {package}`'
    return PantherError(msg)


def cli_error(message: str | Exception) -> None:
    logger.error(message)
    logger.info('Use "panther -h" for more help')


def cli_warning(message: str | Exception, hint: str = None) -> None:
    logger.warning(message)
    if hint:
        logger.info(hint)


def cli_info(message: str) -> None:
    logger.info(message)
    logger.info('Use "panther -h" for more help\n')


def clean_args(args: list[str]) -> dict:
    """
    Input: ['--reload', '--host', '127.0.0.1', ...]
    Output: {'--reload: None, 'host': '127.0.0.1', ...}
    """
    _args = {}
    for i, arg in enumerate(args):
        if arg.startswith('--'):
            if (i + 1) < len(args):
                _args[arg[2:]] = args[i + 1]
            else:
                _args[arg[2:]] = True
    return _args


def print_help_message():
    rprint(help_message)


def print_uvicorn_help_message():
    rprint('Run `uvicorn --help` for more help')


def print_info(config: Config):
    from panther.db.connections import redis

    mo = config.MONITORING
    lq = config.LOG_QUERIES
    bt = config.BACKGROUND_TASKS
    ws = config.HAS_WS
    rd = redis.is_connected
    bd = '{0:<39}'.format(str(config.BASE_DIR))
    if len(bd) > 39:
        bd = f'{bd[:36]}...'

    # Monitoring
    if config.MONITORING:
        monitor = f'{h} * Run "panther monitor" in another session for Monitoring{h}\n'
    else:
        monitor = None

    # Uvloop
    uvloop_msg = None
    if platform.system() != 'Windows':
        try:
            import uvloop
        except ImportError:
            uvloop_msg = (
                f'{h} * You may want to install `uvloop` for better performance{h}\n'
                f'{h}   `pip install uvloop`                                   {h}\n')

    # Gunicorn if Websocket
    gunicorn_msg = None
    if config.HAS_WS:
        try:
            import gunicorn
            gunicorn_msg = f'{h} * You have WS so make sure to run gunicorn with --preload{h}\n'
        except ImportError:
            pass

    # Message
    info_message = f"""{logo}
{h}   Redis: {rd}                                       \t   {h}
{h}   Websocket: {ws}                                   \t   {h}
{h}   Monitoring: {mo}                                  \t   {h}
{h}   Log Queries: {lq}                                 \t   {h}
{h}   Background Tasks: {bt}                            \t   {h}
{h}   Base directory: {bd}{h}
"""
    if monitor:
        info_message += monitor
    if uvloop_msg:
        info_message += uvloop_msg
    if gunicorn_msg:
        info_message += gunicorn_msg

    info_message += bottom
    rprint(info_message)



import asyncio
from datetime import timedelta
from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API
from panther.test import APIClient
from panther.throttling import Throttling


@API()
async def without_throttling_api():
    return 'ok'


@API(throttling=Throttling(rate=3, duration=timedelta(seconds=3)))
async def with_throttling_api():
    return 'ok'


urls = {
    'without-throttling': without_throttling_api,
    'with-throttling': with_throttling_api,
}


class TestThrottling(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    async def test_without_throttling(self):
        res1 = await self.client.get('without-throttling')
        assert res1.status_code == 200

        res2 = await self.client.get('without-throttling')
        assert res2.status_code == 200

        res3 = await self.client.get('without-throttling')
        assert res3.status_code == 200

    async def test_with_throttling(self):
        res1 = await self.client.get('with-throttling')
        assert res1.status_code == 200

        res2 = await self.client.get('with-throttling')
        assert res2.status_code == 200

        res3 = await self.client.get('with-throttling')
        assert res3.status_code == 200

        res4 = await self.client.get('with-throttling')
        assert res4.status_code == 429

        res5 = await self.client.get('with-throttling')
        assert res5.status_code == 429

        await asyncio.sleep(3)  # Sleep and try again

        res6 = await self.client.get('with-throttling')
        assert res6.status_code == 200

        res7 = await self.client.get('with-throttling')
        assert res7.status_code == 200

        res8 = await self.client.get('with-throttling')
        assert res8.status_code == 200

        res9 = await self.client.get('with-throttling')
        assert res9.status_code == 429

        res10 = await self.client.get('with-throttling')
        assert res10.status_code == 429



import asyncio
from types import NoneType
from typing import Generator, AsyncGenerator, Any, Type

import orjson as json
from pydantic import BaseModel

from panther import status
from panther._utils import to_async_generator
from panther.db.cursor import Cursor
from pantherdb import Cursor as PantherDBCursor
from panther.monitoring import Monitoring
from panther.pagination import Pagination

ResponseDataTypes = list | tuple | set | Cursor | PantherDBCursor | dict | int | float | str | bool | bytes | NoneType | Type[BaseModel]
IterableDataTypes = list | tuple | set | Cursor | PantherDBCursor
StreamingDataTypes = Generator | AsyncGenerator


class Response:
    content_type = 'application/json'

    def __init__(
        self,
        data: ResponseDataTypes = None,
        headers: dict | None = None,
        status_code: int = status.HTTP_200_OK,
        pagination: Pagination | None = None,
    ):
        """
        :param data: should be an instance of ResponseDataTypes
        :param headers: should be dict of headers
        :param status_code: should be int
        :param pagination: instance of Pagination or None
            Its template() method will be used
        """
        self.headers = headers or {}
        self.pagination: Pagination | None = pagination
        if isinstance(data, Cursor):
            data = list(data)
        self.initial_data = data
        self.data = self.prepare_data(data=data)
        self.status_code = self.check_status_code(status_code=status_code)

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data

        if self.data is None:
            return b''
        return json.dumps(self.data)

    @property
    def headers(self) -> dict:
        return {
            'Content-Type': self.content_type,
            'Content-Length': len(self.body),
            'Access-Control-Allow-Origin': '*',
        } | self._headers

    @property
    def bytes_headers(self) -> list[list[bytes]]:
        return [[k.encode(), str(v).encode()] for k, v in (self.headers or {}).items()]

    @headers.setter
    def headers(self, headers: dict):
        self._headers = headers

    def prepare_data(self, data: Any):
        """Make sure the response data is only ResponseDataTypes or Iterable of ResponseDataTypes"""
        if isinstance(data, (int | float | str | bool | bytes | NoneType)):
            return data

        elif isinstance(data, dict):
            return {key: self.prepare_data(value) for key, value in data.items()}

        elif issubclass(type(data), BaseModel):
            return data.model_dump()

        elif isinstance(data, IterableDataTypes):
            return [self.prepare_data(d) for d in data]

        else:
            msg = f'Invalid Response Type: {type(data)}'
            raise TypeError(msg)

    @classmethod
    def check_status_code(cls, status_code: Any):
        if not isinstance(status_code, int):
            error = f'Response `status_code` Should Be `int`. (`{status_code}` is {type(status_code)})'
            raise TypeError(error)
        return status_code

    async def apply_output_model(self, output_model: Type[BaseModel]):
        """This method is called in API.__call__"""

        # Dict
        if isinstance(self.data, dict):
            # Apply `validation_alias` (id -> _id)
            for field_name, field in output_model.model_fields.items():
                if field.validation_alias and field_name in self.data:
                    self.data[field.validation_alias] = self.data.pop(field_name)
            output = output_model(**self.data)
            if hasattr(output_model, 'prepare_response'):
                return await output.prepare_response(instance=self.initial_data, data=output.model_dump())
            return output.model_dump()

        # Iterable
        results = []
        if isinstance(self.data, IterableDataTypes):
            for i, d in enumerate(self.data):
                # Apply `validation_alias` (id -> _id)
                for field_name, field in output_model.model_fields.items():
                    if field.validation_alias and field_name in d:
                        d[field.validation_alias] = d.pop(field_name)

                output = output_model(**d)
                if hasattr(output_model, 'prepare_response'):
                    result = await output.prepare_response(instance=self.initial_data[i], data=output.model_dump())
                else:
                    result = output.model_dump()
                results.append(result)
            return results

        # Str | Bool | Bytes
        msg = 'Type of Response data is not match with `output_model`.\n*hint: You may want to remove `output_model`'
        raise TypeError(msg)

    async def send_headers(self, send, /):
        await send({'type': 'http.response.start', 'status': self.status_code, 'headers': self.bytes_headers})

    async def send_body(self, send, receive, /):
        await send({'type': 'http.response.body', 'body': self.body, 'more_body': False})

    async def send(self, send, receive, /, monitoring: Monitoring):
        await self.send_headers(send)
        await self.send_body(send, receive)
        await monitoring.after(self.status_code)

    def __str__(self):
        if len(data := str(self.data)) > 30:
            data = f'{data:.27}...'
        return f'Response(status_code={self.status_code}, data={data})'

    __repr__ = __str__


class StreamingResponse(Response):
    content_type = 'application/octet-stream'

    def __init__(self, *args, **kwargs):
        self.connection_closed = False
        super().__init__(*args, **kwargs)

    async def listen_to_disconnection(self, receive):
        message = await receive()
        if message['type'] == 'http.disconnect':
            self.connection_closed = True

    def prepare_data(self, data: any) -> AsyncGenerator:
        if isinstance(data, AsyncGenerator):
            return data
        elif isinstance(data, Generator):
            return to_async_generator(data)
        msg = f'Invalid Response Type: {type(data)}'
        raise TypeError(msg)

    @property
    def headers(self) -> dict:
        return {
            'Content-Type': self.content_type,
            'Access-Control-Allow-Origin': '*',
        } | self._headers

    @headers.setter
    def headers(self, headers: dict):
        self._headers = headers

    @property
    async def body(self) -> AsyncGenerator:
        async for chunk in self.data:
            if isinstance(chunk, bytes):
                yield chunk
            elif chunk is None:
                yield b''
            else:
                yield json.dumps(chunk)

    async def send_body(self, send, receive, /):
        asyncio.create_task(self.listen_to_disconnection(receive))
        async for chunk in self.body:
            if self.connection_closed:
                break
            await send({'type': 'http.response.body', 'body': chunk, 'more_body': True})
        else:
            await send({'type': 'http.response.body', 'body': b'', 'more_body': False})


class HTMLResponse(Response):
    content_type = 'text/html; charset=utf-8'

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        return self.data.encode()


class PlainTextResponse(Response):
    content_type = 'text/plain; charset=utf-8'

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        return self.data.encode()



import os
from datetime import datetime, timedelta
from pathlib import Path
from unittest import TestCase

from panther import Panther
from panther.configs import config
from panther.middlewares import BaseMiddleware
from panther.utils import generate_secret_key, generate_hash_value_from_string, load_env, round_datetime


class TestLoadEnvFile(TestCase):
    file_path = 'tests/env-test'

    is_active = True
    db_host = '127.0.0.1'
    db_port = 27017

    def tearDown(self) -> None:
        Path(self.file_path).unlink(missing_ok=True)

    def _create_env_file(self, file_data):
        with open(self.file_path, 'w') as file:
            file.write(file_data)

    def test_load_env_invalid_file(self):
        try:
            load_env('fake.file')
        except ValueError as e:
            assert e.args[0] == '"fake.file" is not a file.'

    def test_load_env_double_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = "{self.is_active}"
DB_HOST = "{self.db_host}"
DB_PORT = "{self.db_port}"
        """)

        variables = load_env(self.file_path)
        assert (variables['IS_ACTIVE'] == 'True') == self.is_active
        assert variables['DB_HOST'] == self.db_host
        assert variables['DB_PORT'] == str(self.db_port)

    def test_load_env_single_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = '{self.is_active}'
DB_HOST = '{self.db_host}'
DB_PORT = '{self.db_port}'
                """)

        variables = load_env(self.file_path)
        assert (variables['IS_ACTIVE'] == 'True') == self.is_active
        assert variables['DB_HOST'] == self.db_host
        assert variables['DB_PORT'] == str(self.db_port)

    def test_load_env_no_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = {self.is_active}
DB_HOST = {self.db_host}
DB_PORT = {self.db_port}

                    """)

        variables = load_env(self.file_path)
        assert (variables['IS_ACTIVE'] == 'True') == self.is_active
        assert variables['DB_HOST'] == self.db_host
        assert variables['DB_PORT'] == str(self.db_port)

    def test_load_env_no_space(self):
        self._create_env_file(f"""
IS_ACTIVE={self.is_active}
DB_HOST={self.db_host}
DB_PORT={self.db_port}
                    """)

        variables = load_env(self.file_path)
        assert (variables['IS_ACTIVE'] == 'True') == self.is_active
        assert variables['DB_HOST'] == self.db_host
        assert variables['DB_PORT'] == str(self.db_port)

    def test_load_env_not_striped(self):
        self._create_env_file(f"""
        IS_ACTIVE = {self.is_active}
        DB_HOST = {self.db_host}
        DB_PORT = {self.db_port}
                    """)

        variables = load_env(self.file_path)
        assert (variables['IS_ACTIVE'] == 'True') == self.is_active
        assert variables['DB_HOST'] == self.db_host
        assert variables['DB_PORT'] == str(self.db_port)

    def test_load_env_and_read_from_system_env(self):
        self._create_env_file(f"""
IS_ACTIVE = '{self.is_active}'
DB_HOST = '{self.db_host}'
DB_PORT = '{self.db_port}'
                """)

        load_env(self.file_path)
        assert (os.environ['IS_ACTIVE'] == 'True') == self.is_active
        assert os.environ['DB_HOST'] == self.db_host
        assert os.environ['DB_PORT'] == str(self.db_port)


class TestUtilFunctions(TestCase):

    def test_round_datetime_second_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(seconds=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=40)
        assert rounded_datetime == expected_datetime

    def test_round_datetime_second_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=35)
        _delta = timedelta(seconds=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=40)
        assert rounded_datetime == expected_datetime

    def test_round_datetime_minute_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=15, second=30)
        _delta = timedelta(minutes=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=20)
        assert rounded_datetime == expected_datetime

    def test_round_datetime_minute_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(minutes=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=20)
        assert rounded_datetime == expected_datetime

    def test_round_datetime_hour_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(hours=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12)
        assert rounded_datetime == expected_datetime

    def test_round_datetime_hour_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=10, minute=10, second=30)
        _delta = timedelta(hours=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12)
        assert rounded_datetime == expected_datetime

    def test_round_datetime_day_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=10, minute=10, second=30)
        _delta = timedelta(days=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=23)
        assert rounded_datetime == expected_datetime

    def test_round_datetime_day_2(self):
        _datetime = datetime(year=1997, month=12, day=22, hour=10, minute=10, second=30)
        _delta = timedelta(days=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=23)
        assert rounded_datetime == expected_datetime

    def test_generate_hash_value_from_string(self):
        text = 'Hello World'
        hashed_1 = generate_hash_value_from_string(text)
        hashed_2 = generate_hash_value_from_string(text)

        assert hashed_1 == hashed_2
        assert text != hashed_1


class TestLoadConfigs(TestCase):
    def setUp(self):
        config.refresh()

    def test_urls_not_found(self):
        global URLs
        URLs = None

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'URLs': required."

    def test_urls_cant_be_dict(self):
        global URLs
        URLs = {}

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        msg = (
            "Invalid 'URLs': can't be 'dict', you may want to pass it's value directly to Panther(). "
            "Example: Panther(..., urls=...)"
        )
        assert captured.records[0].getMessage() == msg

    def test_urls_not_string(self):
        global URLs
        URLs = True

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'URLs': should be dotted string."

    def test_urls_invalid_target(self):
        global URLs
        URLs = 'tests.test_utils.TestLoadConfigs'

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'URLs': should point to a dict."

    def test_urls_invalid_module_path(self):
        global URLs
        URLs = 'fake.module'

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'URLs': No module named 'fake'"

    def test_middlewares_invalid_path(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            ('fake.module', {})
        ]

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'MIDDLEWARES': fake.module is not a valid middleware path"

    def test_middlewares_invalid_structure(self):
        global MIDDLEWARES
        MIDDLEWARES = ['fake.module']

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'MIDDLEWARES': fake.module should have 2 part: (path, kwargs)"

    def test_middlewares_too_many_args(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            ('fake.module', 1, 2)
        ]

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'MIDDLEWARES': ('fake.module', 1, 2) too many arguments"

    def test_middlewares_without_args(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            ('tests.test_utils.CorrectTestMiddleware', )
        ]

        with self.assertNoLogs(level='ERROR'):
            Panther(name=__name__, configs=__name__, urls={})

        MIDDLEWARES = []

    def test_middlewares_invalid_middleware_parent(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            ('tests.test_utils.TestMiddleware', )
        ]

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'MIDDLEWARES': is not a sub class of BaseMiddleware"

    def test_jwt_auth_without_secret_key(self):
        global AUTHENTICATION
        AUTHENTICATION = 'panther.authentications.JWTAuthentication'

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                AUTHENTICATION = None

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'JWTConfig': `JWTConfig.key` or `SECRET_KEY` is required."

    def test_jwt_auth_with_secret_key(self):
        global AUTHENTICATION, SECRET_KEY
        AUTHENTICATION = 'panther.authentications.JWTAuthentication'
        SECRET_KEY = generate_secret_key()

        with self.assertNoLogs(level='ERROR'):
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert False
            else:
                assert True
            finally:
                AUTHENTICATION = None
                SECRET_KEY = None


class CorrectTestMiddleware(BaseMiddleware):
    pass


class TestMiddleware:
    pass



from collections import namedtuple
from collections.abc import Callable
from urllib.parse import parse_qsl

from panther.db import Model
from panther.exceptions import InvalidPathVariableAPIError


class Headers:
    accept: str
    accept_encoding: str
    accept_language: str
    authorization: str
    cache_control: str
    connection: str
    content_length: str
    content_type: str
    host: str
    origin: str
    pragma: str
    referer: str
    sec_fetch_dest: str
    sec_fetch_mode: str
    sec_fetch_site: str
    user_agent: str

    upgrade: str
    sec_websocket_version: str
    sec_websocket_key: str

    def __init__(self, headers: list):
        self.__headers = {header[0].decode('utf-8'): header[1].decode('utf-8') for header in headers}
        self.__pythonic_headers = {k.lower().replace('-', '_'): v for k, v in self.__headers.items()}

    def __getattr__(self, item: str):
        if result := self.__pythonic_headers.get(item):
            return result
        return self.__headers.get(item)

    def __getitem__(self, item: str):
        if result := self.__headers.get(item):
            return result
        return self.__pythonic_headers.get(item)

    def __str__(self):
        items = ', '.join(f'{k}={v}' for k, v in self.__headers.items())
        return f'Headers({items})'

    __repr__ = __str__

    @property
    def __dict__(self):
        return self.__headers


Address = namedtuple('Address', ['ip', 'port'])


class BaseRequest:
    def __init__(self, scope: dict, receive: Callable, send: Callable):
        self.scope = scope
        self.asgi_send = send
        self.asgi_receive = receive
        self._headers: Headers | None = None
        self._params: dict | None = None
        self.user: Model | None = None
        self.path_variables: dict | None = None

    @property
    def headers(self) -> Headers:
        if self._headers is None:
            self._headers = Headers(self.scope['headers'])
        return self._headers

    @property
    def query_params(self) -> dict:
        if self._params is None:
            self._params = {k: v for k, v in parse_qsl(self.scope['query_string'].decode('utf-8'))}
        return self._params

    @property
    def path(self) -> str:
        return self.scope['path']

    @property
    def server(self) -> Address:
        return Address(*self.scope['server'])

    @property
    def client(self) -> Address:
        return Address(*self.scope['client'])

    @property
    def http_version(self) -> str:
        return self.scope['http_version']

    @property
    def scheme(self) -> str:
        return self.scope['scheme']

    def collect_path_variables(self, found_path: str):
        self.path_variables = {
            variable.strip('< >'): value
            for variable, value in zip(
                found_path.strip('/').split('/'),
                self.path.strip('/').split('/')
            )
            if variable.startswith('<')
        }

    def clean_parameters(self, func: Callable) -> dict:
        kwargs = self.path_variables.copy()

        for variable_name, variable_type in func.__annotations__.items():
            # Put Request/ Websocket In kwargs (If User Wants It)
            if issubclass(variable_type, BaseRequest):
                kwargs[variable_name] = self

            elif variable_name in kwargs:
                # Cast To Boolean
                if variable_type is bool:
                    kwargs[variable_name] = kwargs[variable_name].lower() not in ['false', '0']

                # Cast To Int
                elif variable_type is int:
                    try:
                        kwargs[variable_name] = int(kwargs[variable_name])
                    except ValueError:
                        raise InvalidPathVariableAPIError(value=kwargs[variable_name], variable_type=variable_type)
        return kwargs



import asyncio
import importlib
import logging
import re
import subprocess
import types
from typing import Any, Generator, Iterator, AsyncGenerator
from collections.abc import Callable
from traceback import TracebackException

from panther.exceptions import PantherError
from panther.file_handler import File

logger = logging.getLogger('panther')


def import_class(dotted_path: str, /) -> type[Any]:
    """
    Example:
    -------
        Input: panther.db.models.User
        Output: User (The Class)
    """
    path, name = dotted_path.rsplit('.', 1)
    module = importlib.import_module(path)
    return getattr(module, name)


def read_multipart_form_data(boundary: str, body: bytes) -> dict:
    boundary = b'--' + boundary.encode()
    new_line = b'\r\n' if body[-2:] == b'\r\n' else b'\n'

    field_pattern = (
            rb'(Content-Disposition: form-data; name=")(.*)("'
            + 2 * new_line
            + b')(.*)'
    )
    file_pattern = (
            rb'(Content-Disposition: form-data; name=")(.*)("; filename=")(.*)("'
            + new_line
            + b'Content-Type: )(.*)'
    )

    data = {}
    for _row in body.split(boundary):
        row = _row.removeprefix(new_line).removesuffix(new_line)

        if row in (b'', b'--'):
            continue

        if match := re.match(pattern=field_pattern, string=row):
            _, field_name, _, value = match.groups()
            data[field_name.decode('utf-8')] = value.decode('utf-8')

        else:
            file_meta_data, value = row.split(2 * new_line, 1)

            if match := re.match(pattern=file_pattern, string=file_meta_data):
                _, field_name, _, file_name, _, content_type = match.groups()
                file = File(
                    file_name=file_name.decode('utf-8'),
                    content_type=content_type.decode('utf-8'),
                    file=value,
                )
                data[field_name.decode('utf-8')] = file
            else:
                logger.error('Unrecognized Pattern')

    return data


def is_function_async(func: Callable) -> bool:
    """
    Sync result is 0 --> False
    async result is 128 --> True
    """
    return bool(func.__code__.co_flags & (1 << 7))


def clean_traceback_message(exception: Exception) -> str:
    """We are ignoring packages traceback message"""
    tb = TracebackException(type(exception), exception, exception.__traceback__)
    stack = tb.stack.copy()
    for t in stack:
        if t.filename.find('site-packages/panther') != -1 or t.filename.find('site-packages\\panther') != -1:
            tb.stack.remove(t)
    _traceback = list(tb.format(chain=False))
    return exception if len(_traceback) == 1 else f'{exception}\n' + ''.join(_traceback)


def reformat_code(base_dir):
    try:
        subprocess.run(['ruff', 'format', base_dir])
        subprocess.run(['ruff', 'check', '--select', 'I', '--fix', base_dir])
    except FileNotFoundError:
        raise PantherError("No module named 'ruff', Hint: `pip install ruff`")


def check_function_type_endpoint(endpoint: types.FunctionType) -> Callable:
    # Function Doesn't Have @API Decorator
    if not hasattr(endpoint, '__wrapped__'):
        logger.critical(f'You may have forgotten to use @API() on the {endpoint.__name__}()')
        raise TypeError
    return endpoint


def check_class_type_endpoint(endpoint: Callable) -> Callable:
    from panther.app import GenericAPI

    if not issubclass(endpoint, GenericAPI):
        logger.critical(f'You may have forgotten to inherit from GenericAPI on the {endpoint.__name__}()')
        raise TypeError

    return endpoint().call_method


def async_next(iterator: Iterator):
    """
    The StopIteration exception is a special case in Python,
    particularly when it comes to asynchronous programming and the use of asyncio.
    This is because StopIteration is not meant to be caught in the traditional sense;
        it's used internally by Python to signal the end of an iteration.
    """
    try:
        return next(iterator)
    except StopIteration:
        raise StopAsyncIteration


async def to_async_generator(generator: Generator) -> AsyncGenerator:
    while True:
        try:
            yield await asyncio.to_thread(async_next, iter(generator))
        except StopAsyncIteration:
            break



import time
import asyncio
from datetime import timedelta
from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API
from panther.test import APIClient
from tests._utils import check_two_dicts


@API()
async def without_cache_api():
    await asyncio.sleep(0.01)
    return {'detail': time.time()}


@API(cache=True)
async def with_cache_api():
    await asyncio.sleep(0.01)
    return {'detail': time.time()}


@API(cache=True, cache_exp_time=timedelta(seconds=5))
async def expired_cache_api():
    await asyncio.sleep(0.01)
    return {'detail': time.time()}


urls = {
    'without-cache': without_cache_api,
    'with-cache': with_cache_api,
    'with-expired-cache': expired_cache_api,
}


class TestCaching(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    async def test_without_cache(self):
        res1 = await self.client.get('without-cache')
        assert res1.status_code == 200

        res2 = await self.client.get('without-cache')
        assert res2.status_code == 200

        assert check_two_dicts(res1.data, res2.data) is False

    async def test_with_cache(self):
        res1 = await self.client.get('with-cache')
        assert res1.status_code == 200

        res2 = await self.client.get('with-cache')
        assert res2.status_code == 200

        assert check_two_dicts(res1.data, res2.data) is True

    async def test_with_cache_5second_exp_time(self):
        # First Request
        with self.assertLogs(level='INFO') as captured:
            res1 = await self.client.get('with-expired-cache')
        assert res1.status_code == 200

        # Check Logs
        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == '`cache_exp_time` is not very accurate when `redis` is not connected.'

        # Second Request
        res2 = await self.client.get('with-expired-cache')
        assert res2.status_code == 200

        # Response should be cached
        assert check_two_dicts(res1.data, res2.data) is True

        await asyncio.sleep(5)

        # Third Request
        res3 = await self.client.get('with-expired-cache')
        assert res3.status_code == 200

        # After 5 seconds we should have a new response
        assert check_two_dicts(res1.data, res3.data) is False




from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API, GenericAPI
from panther.response import Response, HTMLResponse, PlainTextResponse, StreamingResponse
from panther.test import APIClient


@API()
async def return_nothing():
    pass


class ReturnNothing(GenericAPI):
    def get(self):
        pass


@API()
async def return_none():
    return None


class ReturnNone(GenericAPI):
    def get(self):
        return None


@API()
async def return_string():
    return 'Hello'


class ReturnString(GenericAPI):
    def get(self):
        return 'Hello'


@API()
async def return_dict():
    return {'detail': 'ok'}


class ReturnDict(GenericAPI):
    def get(self):
        return {'detail': 'ok'}


@API()
async def return_list():
    return [1, 2, 3]


class ReturnList(GenericAPI):
    def get(self):
        return [1, 2, 3]


@API()
async def return_tuple():
    return 1, 2, 3, 4


class ReturnTuple(GenericAPI):
    def get(self):
        return 1, 2, 3, 4


@API()
async def return_response_none():
    return Response()


class ReturnResponseNone(GenericAPI):
    def get(self):
        return Response()


@API()
async def return_response_dict():
    return Response(data={'detail': 'ok'})


class ReturnResponseDict(GenericAPI):
    def get(self):
        return Response(data={'detail': 'ok'})


@API()
async def return_response_list():
    return Response(data=['car', 'home', 'phone'])


class ReturnResponseList(GenericAPI):
    def get(self):
        return Response(data=['car', 'home', 'phone'])


@API()
async def return_response_tuple():
    return Response(data=('car', 'home', 'phone', 'book'))


class ReturnResponseTuple(GenericAPI):
    def get(self):
        return Response(data=('car', 'home', 'phone', 'book'))


@API()
async def return_html_response():
    return HTMLResponse('<html><head><title></title></head></html>')


class ReturnHTMLResponse(GenericAPI):
    def get(self):
        return HTMLResponse('<html><head><title></title></head></html>')


@API()
async def return_plain_response():
    return PlainTextResponse('Hello World')


class ReturnPlainResponse(GenericAPI):
    def get(self):
        return PlainTextResponse('Hello World')


class ReturnStreamingResponse(GenericAPI):
    def get(self):
        def f():
            for i in range(5):
                yield i
        return StreamingResponse(f())


class ReturnAsyncStreamingResponse(GenericAPI):
    async def get(self):
        async def f():
            for i in range(6):
                yield i
        return StreamingResponse(f())


class ReturnInvalidStatusCode(GenericAPI):
    def get(self):
        return Response(status_code='ali')


urls = {
    'nothing': return_nothing,
    'none': return_none,
    'dict': return_dict,
    'str': return_string,
    'list': return_list,
    'tuple': return_tuple,
    'response-none': return_response_none,
    'response-dict': return_response_dict,
    'response-list': return_response_list,
    'response-tuple': return_response_tuple,
    'html': return_html_response,
    'plain': return_plain_response,

    'nothing-cls': ReturnNothing,
    'none-cls': ReturnNone,
    'dict-cls': ReturnDict,
    'str-cls': ReturnString,
    'list-cls': ReturnList,
    'tuple-cls': ReturnTuple,
    'response-none-cls': ReturnResponseNone,
    'response-dict-cls': ReturnResponseDict,
    'response-list-cls': ReturnResponseList,
    'response-tuple-cls': ReturnResponseTuple,
    'html-cls': ReturnHTMLResponse,
    'plain-cls': ReturnPlainResponse,

    'stream': ReturnStreamingResponse,
    'async-stream': ReturnAsyncStreamingResponse,
    'invalid-status-code': ReturnInvalidStatusCode,
}


class TestResponses(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    async def test_nothing(self):
        res = await self.client.get('nothing/')
        assert res.status_code == 200
        assert res.data is None
        assert res.body == b''
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '0'

    async def test_nothing_cls(self):
        res = await self.client.get('nothing-cls/')
        assert res.status_code == 200
        assert res.data is None
        assert res.body == b''
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '0'

    async def test_none(self):
        res = await self.client.get('none/')
        assert res.status_code == 200
        assert res.data is None
        assert res.body == b''
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '0'

    async def test_none_cls(self):
        res = await self.client.get('none-cls/')
        assert res.status_code == 200
        assert res.data is None
        assert res.body == b''
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '0'

    async def test_dict(self):
        res = await self.client.get('dict/')
        assert res.status_code == 200
        assert res.data == {'detail': 'ok'}
        assert res.body == b'{"detail":"ok"}'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '15'

    async def test_dict_cls(self):
        res = await self.client.get('dict-cls/')
        assert res.status_code == 200
        assert res.data == {'detail': 'ok'}
        assert res.body == b'{"detail":"ok"}'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '15'

    async def test_string(self):
        res = await self.client.get('str/')
        assert res.status_code == 200
        assert res.data == 'Hello'
        assert res.body == b'"Hello"'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '7'

    async def test_string_cls(self):
        res = await self.client.get('str-cls/')
        assert res.status_code == 200
        assert res.data == 'Hello'
        assert res.body == b'"Hello"'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '7'

    async def test_list(self):
        res = await self.client.get('list/')
        assert res.status_code == 200
        assert res.data == [1, 2, 3]
        assert res.body == b'[1,2,3]'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '7'

    async def test_list_cls(self):
        res = await self.client.get('list-cls/')
        assert res.status_code == 200
        assert res.data == [1, 2, 3]
        assert res.body == b'[1,2,3]'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '7'

    async def test_tuple(self):
        res = await self.client.get('tuple/')
        assert res.status_code == 200
        assert res.data == [1, 2, 3, 4]
        assert res.body == b'[1,2,3,4]'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '9'

    async def test_tuple_cls(self):
        res = await self.client.get('tuple-cls/')
        assert res.status_code == 200
        assert res.data == [1, 2, 3, 4]
        assert res.body == b'[1,2,3,4]'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '9'

    async def test_response_none(self):
        res = await self.client.get('response-none/')
        assert res.status_code == 200
        assert res.data is None
        assert res.body == b''
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '0'

    async def test_response_none_cls(self):
        res = await self.client.get('response-none-cls/')
        assert res.status_code == 200
        assert res.data is None
        assert res.body == b''
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '0'

    async def test_response_dict(self):
        res = await self.client.get('response-dict/')
        assert res.status_code == 200
        assert res.data == {'detail': 'ok'}
        assert res.body == b'{"detail":"ok"}'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '15'

    async def test_response_dict_cls(self):
        res = await self.client.get('response-dict-cls/')
        assert res.status_code == 200
        assert res.data == {'detail': 'ok'}
        assert res.body == b'{"detail":"ok"}'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '15'

    async def test_response_list(self):
        res = await self.client.get('response-list/')
        assert res.status_code == 200
        assert res.data == ['car', 'home', 'phone']
        assert res.body == b'["car","home","phone"]'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '22'

    async def test_response_list_cls(self):
        res = await self.client.get('response-list-cls/')
        assert res.status_code == 200
        assert res.data == ['car', 'home', 'phone']
        assert res.body == b'["car","home","phone"]'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '22'

    async def test_response_tuple(self):
        res = await self.client.get('response-tuple/')
        assert res.status_code == 200
        assert res.data == ['car', 'home', 'phone', 'book']
        assert res.body == b'["car","home","phone","book"]'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '29'

    async def test_response_tuple_cls(self):
        res = await self.client.get('response-tuple-cls/')
        assert res.status_code == 200
        assert res.data == ['car', 'home', 'phone', 'book']
        assert res.body == b'["car","home","phone","book"]'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '29'

    async def test_response_html(self):
        res = await self.client.get('html/')
        assert res.status_code == 200
        assert res.data == '<html><head><title></title></head></html>'
        assert res.body == b'<html><head><title></title></head></html>'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '41'

    async def test_response_html_cls(self):
        res = await self.client.get('html-cls/')
        assert res.status_code == 200
        assert res.data == '<html><head><title></title></head></html>'
        assert res.body == b'<html><head><title></title></head></html>'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '41'

    async def test_response_plain(self):
        res = await self.client.get('plain/')
        assert res.status_code == 200
        assert res.data == 'Hello World'
        assert res.body == b'Hello World'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'text/plain; charset=utf-8'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '11'

    async def test_response_plain_cls(self):
        res = await self.client.get('plain-cls/')
        assert res.status_code == 200
        assert res.data == 'Hello World'
        assert res.body == b'Hello World'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'text/plain; charset=utf-8'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == '11'

    async def test_streaming_response(self):
        res = await self.client.get('stream/')
        assert res.status_code == 200
        assert res.headers['Content-Type'] == 'application/octet-stream'
        assert res.data == '01234'
        assert res.body == b'01234'

    async def test_async_streaming_response(self):
        res = await self.client.get('async-stream/')
        assert res.status_code == 200
        assert res.headers['Content-Type'] == 'application/octet-stream'
        assert res.data == '012345'
        assert res.body == b'012345'

    async def test_invalid_status_code(self):
        with self.assertLogs(level='CRITICAL') as captured:
            res = await self.client.get('invalid-status-code/')

        assert len(captured.records) == 1
        assert captured.records[0].getMessage().split('\n')[0] == "Response `status_code` Should Be `int`. (`ali` is <class 'str'>)"

        assert res.status_code == 500
        assert res.data == {'detail': 'Internal Server Error'}
        assert res.body == b'{"detail":"Internal Server Error"}'
        assert set(res.headers.keys()) == {'Content-Type', 'Access-Control-Allow-Origin', 'Content-Length'}
        assert res.headers['Content-Type'] == 'application/json'
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Content-Length'] == 34



import functools
import logging
from datetime import timedelta
from typing import Literal

from orjson import JSONDecodeError
from pydantic import ValidationError, BaseModel

from panther._utils import is_function_async
from panther.caching import (
    get_response_from_cache,
    set_response_in_cache,
    get_throttling_from_cache,
    increment_throttling_in_cache
)
from panther.configs import config
from panther.exceptions import (
    APIError,
    AuthorizationAPIError,
    JSONDecodeAPIError,
    MethodNotAllowedAPIError,
    ThrottlingAPIError,
    BadRequestAPIError
)
from panther.request import Request
from panther.response import Response
from panther.serializer import ModelSerializer
from panther.throttling import Throttling

__all__ = ('API', 'GenericAPI')

logger = logging.getLogger('panther')


class API:
    def __init__(
            self,
            *,
            input_model: type[ModelSerializer] | type[BaseModel] | None = None,
            output_model: type[ModelSerializer] | type[BaseModel] | None = None,
            auth: bool = False,
            permissions: list | None = None,
            throttling: Throttling | None = None,
            cache: bool = False,
            cache_exp_time: timedelta | int | None = None,
            methods: list[Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE']] | None = None,
    ):
        self.input_model = input_model
        self.output_model = output_model
        self.auth = auth
        self.permissions = permissions or []
        self.throttling = throttling
        self.cache = cache
        self.cache_exp_time = cache_exp_time
        self.methods = methods
        self.request: Request | None = None

    def __call__(self, func):
        @functools.wraps(func)
        async def wrapper(request: Request) -> Response:
            self.request = request

            # 0. Preflight
            if self.request.method == 'OPTIONS':
                return self.options()

            # 1. Check Method
            if self.methods and self.request.method not in self.methods:
                raise MethodNotAllowedAPIError

            # 2. Authentication
            await self.handle_authentication()

            # 3. Permissions
            await self.handle_permission()

            # 4. Throttling
            await self.handle_throttling()

            # 5. Validate Input
            if self.request.method in ['POST', 'PUT', 'PATCH']:
                self.handle_input_validation()

            # 6. Get Cached Response
            if self.cache and self.request.method == 'GET':
                if cached := await get_response_from_cache(request=self.request, cache_exp_time=self.cache_exp_time):
                    return Response(data=cached.data, status_code=cached.status_code)

            # 7. Put PathVariables and Request(If User Wants It) In kwargs
            kwargs = self.request.clean_parameters(func)

            # 8. Call Endpoint
            if is_function_async(func):
                response = await func(**kwargs)
            else:
                response = func(**kwargs)

            # 9. Clean Response
            if not isinstance(response, Response):
                response = Response(data=response)
            if self.output_model and response.data:
                response.data = await response.apply_output_model(output_model=self.output_model)
            if response.pagination:
                response.data = await response.pagination.template(response.data)

            # 10. Set New Response To Cache
            if self.cache and self.request.method == 'GET':
                await set_response_in_cache(
                    request=self.request,
                    response=response,
                    cache_exp_time=self.cache_exp_time
                )

            # 11. Warning CacheExpTime
            if self.cache_exp_time and self.cache is False:
                logger.warning('"cache_exp_time" won\'t work while "cache" is False')

            return response

        return wrapper

    async def handle_authentication(self) -> None:
        if self.auth:
            if not config.AUTHENTICATION:
                logger.critical('"AUTHENTICATION" has not been set in configs')
                raise APIError
            self.request.user = await config.AUTHENTICATION.authentication(self.request)

    async def handle_throttling(self) -> None:
        if throttling := self.throttling or config.THROTTLING:
            if await get_throttling_from_cache(self.request, duration=throttling.duration) + 1 > throttling.rate:
                raise ThrottlingAPIError

            await increment_throttling_in_cache(self.request, duration=throttling.duration)

    async def handle_permission(self) -> None:
        for perm in self.permissions:
            if type(perm.authorization).__name__ != 'method':
                logger.error(f'{perm.__name__}.authorization should be "classmethod"')
                raise AuthorizationAPIError
            if await perm.authorization(self.request) is False:
                raise AuthorizationAPIError

    def handle_input_validation(self):
        if self.input_model:
            self.request.validated_data = self.validate_input(model=self.input_model, request=self.request)

    @classmethod
    def options(cls):
        headers = {
            'Access-Control-Allow-Methods': 'DELETE, GET, PATCH, POST, PUT, OPTIONS',
            'Access-Control-Allow-Headers': 'Accept, Authorization, User-Agent, Content-Type',
        }
        return Response(headers=headers)

    @classmethod
    def validate_input(cls, model, request: Request):
        if isinstance(request.data, bytes):
            raise BadRequestAPIError(detail='Content-Type is not valid')
        if request.data is None:
            raise BadRequestAPIError(detail='Request body is required')
        try:
            # `request` will be ignored in regular `BaseModel`
            return model(**request.data, request=request)
        except ValidationError as validation_error:
            error = {'.'.join(str(loc) for loc in e['loc']): e['msg'] for e in validation_error.errors()}
            raise BadRequestAPIError(detail=error)
        except JSONDecodeError:
            raise JSONDecodeAPIError


class GenericAPI:
    input_model: type[ModelSerializer] | type[BaseModel] | None = None
    output_model: type[ModelSerializer] | type[BaseModel] | None = None
    auth: bool = False
    permissions: list | None = None
    throttling: Throttling | None = None
    cache: bool = False
    cache_exp_time: timedelta | int | None = None

    async def get(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def post(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def put(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def patch(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def delete(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def get_input_model(self, request: Request) -> type[ModelSerializer] | type[BaseModel] | None:
        return None

    async def get_output_model(self, request: Request) -> type[ModelSerializer] | type[BaseModel] | None:
        return None

    async def call_method(self, request: Request):
        match request.method:
            case 'GET':
                func = self.get
            case 'POST':
                func = self.post
            case 'PUT':
                func = self.put
            case 'PATCH':
                func = self.patch
            case 'DELETE':
                func = self.delete
            case 'OPTIONS':
                func = API.options
            case _:
                raise MethodNotAllowedAPIError

        return await API(
            input_model=self.input_model or await self.get_input_model(request=request),
            output_model=self.output_model or await self.get_output_model(request=request),
            auth=self.auth,
            permissions=self.permissions,
            throttling=self.throttling,
            cache=self.cache,
            cache_exp_time=self.cache_exp_time,
        )(func)(request=request)



import asyncio

from panther._utils import is_function_async
from panther.configs import config


class Event:
    @staticmethod
    def startup(func):
        config.STARTUPS.append(func)

        def wrapper():
            return func()
        return wrapper

    @staticmethod
    def shutdown(func):
        config.SHUTDOWNS.append(func)

        def wrapper():
            return func()
        return wrapper

    @staticmethod
    async def run_startups():
        for func in config.STARTUPS:
            if is_function_async(func):
                await func()
            else:
                func()

    @staticmethod
    def run_shutdowns():
        for func in config.SHUTDOWNS:
            if is_function_async(func):
                try:
                    asyncio.run(func())
                except ModuleNotFoundError:
                    # Error: import of asyncio halted; None in sys.modules
                    #   And as I figured it out, it only happens when we are running with
                    #   gunicorn and Uvicorn workers (-k uvicorn.workers.UvicornWorker)
                    pass
            else:
                func()



from pathlib import Path
from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.test import APIClient

DB_PATH = 'test.pdb'
DATABASE = {
    'engine': {
        'class': 'panther.db.connections.PantherDBConnection',
        'path': DB_PATH,
    },
}


class TestPanelAPIs(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls={})
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        Path(DB_PATH).unlink()

    async def test_list_of_models(self):
        response = await self.client.get('_panel')
        expected_keys = ['name', 'module', 'index']
        assert expected_keys == [*response.data[0].keys()]



import logging
import time
from abc import abstractmethod
from datetime import timezone, datetime
from typing import Literal

from panther.base_websocket import Websocket
from panther.cli.utils import import_error
from panther.configs import config
from panther.db.connections import redis
from panther.db.models import BaseUser, Model
from panther.exceptions import AuthenticationAPIError
from panther.request import Request
from panther.utils import generate_hash_value_from_string

try:
    from jose import JWTError, jwt
except ModuleNotFoundError as e:
    raise import_error(e, package='python-jose')

logger = logging.getLogger('panther')


class BaseAuthentication:
    @classmethod
    @abstractmethod
    async def authentication(cls, request: Request | Websocket):
        """Return Instance of User"""
        msg = f'{cls.__name__}.authentication() is not implemented.'
        raise cls.exception(msg) from None

    @staticmethod
    def exception(message: str, /) -> type[AuthenticationAPIError]:
        logger.error(f'Authentication Error: "{message}"')
        return AuthenticationAPIError


class JWTAuthentication(BaseAuthentication):
    model = BaseUser
    keyword = 'Bearer'
    algorithm = 'HS256'
    HTTP_HEADER_ENCODING = 'iso-8859-1'  # RFC5987

    @classmethod
    def get_authorization_header(cls, request: Request | Websocket) -> str:
        if auth := request.headers.authorization:
            return auth
        msg = 'Authorization is required'
        raise cls.exception(msg) from None

    @classmethod
    async def authentication(cls, request: Request | Websocket) -> Model:
        auth_header = cls.get_authorization_header(request).split()

        if len(auth_header) != 2:
            msg = 'Authorization should have 2 part'
            raise cls.exception(msg) from None

        bearer, token = auth_header

        try:
            token.encode(JWTAuthentication.HTTP_HEADER_ENCODING)
        except UnicodeEncodeError as e:
            raise cls.exception(e) from None

        if bearer.lower() != cls.keyword.lower():
            msg = 'Authorization keyword is not valid'
            raise cls.exception(msg) from None

        if redis.is_connected and await cls._check_in_cache(token=token):
            msg = 'User logged out'
            raise cls.exception(msg) from None

        payload = cls.decode_jwt(token)
        user = await cls.get_user(payload)
        user._auth_token = token
        return user

    @classmethod
    def decode_jwt(cls, token: str) -> dict:
        """Decode JWT token to user_id (it can return multiple variable ... )"""
        try:
            return jwt.decode(
                token=token,
                key=config.JWT_CONFIG.key,
                algorithms=[config.JWT_CONFIG.algorithm],
            )
        except JWTError as e:
            raise cls.exception(e) from None

    @classmethod
    async def get_user(cls, payload: dict) -> Model:
        """Get UserModel from config, else use default UserModel from cls.model"""
        if (user_id := payload.get('user_id')) is None:
            msg = 'Payload does not have `user_id`'
            raise cls.exception(msg)

        user_model = config.USER_MODEL or cls.model
        if user := await user_model.find_one(id=user_id):
            return user

        msg = 'User not found'
        raise cls.exception(msg) from None

    @classmethod
    def encode_jwt(cls, user_id: str, token_type: Literal['access', 'refresh'] = 'access') -> str:
        """Encode JWT from user_id."""
        issued_at = datetime.now(timezone.utc).timestamp()
        if token_type == 'access':
            expire = issued_at + config.JWT_CONFIG.life_time
        else:
            expire = issued_at + config.JWT_CONFIG.refresh_life_time

        claims = {
            'token_type': token_type,
            'user_id': user_id,
            'iat': issued_at,
            'exp': expire,
        }
        return jwt.encode(
            claims,
            key=config.JWT_CONFIG.key,
            algorithm=config.JWT_CONFIG.algorithm,
        )

    @classmethod
    def login(cls, user_id: str) -> dict:
        """Return dict of access and refresh token"""
        return {
            'access_token': cls.encode_jwt(user_id=user_id),
            'refresh_token': cls.encode_jwt(user_id=user_id, token_type='refresh')
        }

    @classmethod
    async def logout(cls, raw_token: str) -> None:
        *_, token = raw_token.split()
        if redis.is_connected:
            payload = cls.decode_jwt(token=token)
            remaining_exp_time = payload['exp'] - time.time()
            await cls._set_in_cache(token=token, exp=int(remaining_exp_time))
        else:
            logger.error('`redis` middleware is required for `logout()`')

    @classmethod
    async def _set_in_cache(cls, token: str, exp: int) -> None:
        key = generate_hash_value_from_string(token)
        await redis.set(key, b'', ex=exp)

    @classmethod
    async def _check_in_cache(cls, token: str) -> bool:
        key = generate_hash_value_from_string(token)
        return bool(await redis.exists(key))

    @staticmethod
    def exception(message: str | JWTError | UnicodeEncodeError, /) -> type[AuthenticationAPIError]:
        logger.error(f'JWT Authentication Error: "{message}"')
        return AuthenticationAPIError


class QueryParamJWTAuthentication(JWTAuthentication):
    @classmethod
    def get_authorization_header(cls, request: Request | Websocket) -> str:
        if auth := request.query_params.get('authorization'):
            return auth
        msg = 'Authorization is required'
        raise cls.exception(msg) from None



from panther.panel.apis import documents_api, models_api, single_document_api, healthcheck_api

urls = {
    '': models_api,
    '<index>/': documents_api,
    '<index>/<document_id>/': single_document_api,
    'health': healthcheck_api,
}



import contextlib
import os

import uvicorn

from panther.cli.utils import print_uvicorn_help_message, clean_args, cli_error


def _handle_commands(args: dict[str, str | None]) -> dict:
    """
    Boolean Commands:
    - reload
    - access-log
    - no-access-log
    - use-colors
    - no-use-colors
    - server-header
    - no-server-header

    Int Commands:
    - port
    - ws_max_size
    - ws_max_queue
    - ws_ping_interval
    - ws_ping_timeout
    """
    _command = {}

    if 'reload' in args:
        _command['reload'] = bool(args.pop('reload', None))

    _command['access_log'] = False  # Default
    if 'access-log' in args:
        _command['access_log'] = bool(args.pop('access-log', None))

    if 'no-access-log' in args:
        _command['access_log'] = not bool(args.pop('no-access-log', None))

    if 'use-colors' in args:
        _command['use_colors'] = bool(args.pop('use-colors', None))

    if 'no-use-colors' in args:
        _command['use_colors'] = not bool(args.pop('no-use-colors', None))

    if 'server-header' in args:
        _command['server_header'] = bool(args.pop('server-header', None))

    if 'no-server-header' in args:
        _command['server_header'] = not bool(args.pop('no-server-header', None))

    if 'port' in args:
        _command['port'] = int(args.pop('port'))

    if 'ws_max_size' in args:
        _command['ws_max_size'] = int(args.pop('ws_max_size'))

    if 'ws_max_queue' in args:
        _command['ws_max_queue'] = int(args.pop('ws_max_queue'))

    if 'ws_ping_interval' in args:
        _command['ws_ping_interval'] = int(args.pop('ws_ping_interval'))

    if 'ws_ping_timeout' in args:
        _command['ws_ping_timeout'] = int(args.pop('ws_ping_timeout'))

    return _command


def run(args: list[str]) -> None:
    args = clean_args(args)

    if any(a in args for a in ['h', 'help', '-h', '--help']):
        print_uvicorn_help_message()
        return
    command = {'app_dir': os.getcwd()}
    command.update(_handle_commands(args))
    command.update(args)
    try:
        with contextlib.suppress(KeyboardInterrupt):
            uvicorn.run('main:app', **command)
    except TypeError as e:
        cli_error(e)



from __future__ import annotations

from sys import version_info
from typing import Iterable, Sequence

from panther.db.connections import db
from panther.db.cursor import Cursor
from panther.db.queries.base_queries import BaseQuery
from panther.db.utils import prepare_id_for_query

try:
    from bson.codec_options import CodecOptions
except ImportError:
    # This 'CodecOptions' is not going to be used,
    #   If user really wants to use it,
    #   we are going to force him to install it in `panther.db.connections.MongoDBConnection.init`
    CodecOptions = type('CodecOptions', (), {})

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BaseMongoDBQuery')


class BaseMongoDBQuery(BaseQuery):
    @classmethod
    def _merge(cls, *args, is_mongo: bool = True) -> dict:
        return super()._merge(*args, is_mongo=is_mongo)

    # TODO: https://jira.mongodb.org/browse/PYTHON-4192
    # @classmethod
    # def collection(cls):
    #     return db.session.get_collection(name=cls.__name__, codec_options=CodecOptions(document_class=cls))

    # # # # # Find # # # # #
    @classmethod
    async def find_one(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := await db.session[cls.__name__].find_one(cls._merge(_filter, kwargs)):
            return cls._create_model_instance(document=document)
        return None

    @classmethod
    async def find(cls, _filter: dict | None = None, /, **kwargs) -> Cursor:
        return Cursor(cls=cls, collection=db.session[cls.__name__].delegate, filter=cls._merge(_filter, kwargs))

    @classmethod
    async def first(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        cursor = await cls.find(_filter, **kwargs)
        for result in cursor.sort('_id', 1).limit(-1):
            return result
        return None

    @classmethod
    async def last(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        cursor = await cls.find(_filter, **kwargs)
        for result in cursor.sort('_id', -1).limit(-1):
            return result
        return None

    @classmethod
    async def aggregate(cls, pipeline: Sequence[dict]) -> Iterable[dict]:
        return await db.session[cls.__name__].aggregate(pipeline).to_list(None)

    # # # # # Count # # # # #
    @classmethod
    async def count(cls, _filter: dict | None = None, /, **kwargs) -> int:
        return await db.session[cls.__name__].count_documents(cls._merge(_filter, kwargs))

    # # # # # Insert # # # # #
    @classmethod
    async def insert_one(cls, _document: dict | None = None, /, **kwargs) -> Self:
        document = cls._merge(_document, kwargs)
        cls._validate_data(data=document)

        await db.session[cls.__name__].insert_one(document)
        return cls._create_model_instance(document=document)

    @classmethod
    async def insert_many(cls, documents: Iterable[dict]) -> list[Self]:
        for document in documents:
            prepare_id_for_query(document, is_mongo=True)
            cls._validate_data(data=document)

        await db.session[cls.__name__].insert_many(documents)
        return [cls._create_model_instance(document=document) for document in documents]

    # # # # # Delete # # # # #
    async def delete(self) -> None:
        await db.session[self.__class__.__name__].delete_one({'_id': self._id})

    @classmethod
    async def delete_one(cls, _filter: dict | None = None, /, **kwargs) -> bool:
        result = await db.session[cls.__name__].delete_one(cls._merge(_filter, kwargs))
        return bool(result.deleted_count)

    @classmethod
    async def delete_many(cls, _filter: dict | None = None, /, **kwargs) -> int:
        result = await db.session[cls.__name__].delete_many(cls._merge(_filter, kwargs))
        return result.deleted_count

    # # # # # Update # # # # #
    async def update(self, _update: dict | None = None, /, **kwargs) -> None:
        merged_update_query = self._merge(_update, kwargs)
        merged_update_query.pop('_id', None)

        self._validate_data(data=merged_update_query, is_updating=True)

        update_query = {}
        for field, value in merged_update_query.items():
            if field.startswith('$'):
                update_query[field] = value
            else:
                update_query['$set'] = update_query.get('$set', {})
                update_query['$set'][field] = value
                setattr(self, field, value)

        await db.session[self.__class__.__name__].update_one({'_id': self._id}, update_query)

    @classmethod
    async def update_one(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> bool:
        prepare_id_for_query(_filter, is_mongo=True)
        merged_update_query = cls._merge(_update, kwargs)

        update_query = {}
        for field, value in merged_update_query.items():
            if field.startswith('$'):
                update_query[field] = value
            else:
                update_query['$set'] = update_query.get('$set', {})
                update_query['$set'][field] = value

        result = await db.session[cls.__name__].update_one(_filter, update_query)
        return bool(result.matched_count)

    @classmethod
    async def update_many(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> int:
        prepare_id_for_query(_filter, is_mongo=True)
        update_fields = {'$set': cls._merge(_update, kwargs)}

        result = await db.session[cls.__name__].update_many(_filter, update_fields)
        return result.modified_count



import operator
from abc import abstractmethod
from functools import reduce
from sys import version_info
from typing import Iterator

from pydantic_core._pydantic_core import ValidationError

from panther.db.cursor import Cursor
from panther.db.utils import prepare_id_for_query
from panther.exceptions import DatabaseError

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BaseQuery')


class BaseQuery:
    @classmethod
    def _merge(cls, *args, is_mongo: bool = False) -> dict:
        prepare_id_for_query(*args, is_mongo=is_mongo)
        return reduce(operator.ior, filter(None, args), {})

    @classmethod
    def _clean_error_message(cls, validation_error: ValidationError, is_updating: bool = False) -> str:
        error = ', '.join(
            '{field}="{error}"'.format(
                field='.'.join(loc for loc in e['loc']),
                error=e['msg']
            )
            for e in validation_error.errors()
            if not is_updating or e['type'] != 'missing'
        )
        return f'{cls.__name__}({error})' if error else ''

    @classmethod
    def _validate_data(cls, *, data: dict, is_updating: bool = False):
        """Validate document before inserting to collection"""
        try:
            cls(**data)
        except ValidationError as validation_error:
            if error := cls._clean_error_message(validation_error=validation_error, is_updating=is_updating):
                raise DatabaseError(error)

    @classmethod
    def _create_model_instance(cls, document: dict):
        """Prevent getting errors from document insertion"""
        try:
            return cls(**document)
        except ValidationError as validation_error:
            if error := cls._clean_error_message(validation_error=validation_error):
                raise DatabaseError(error)

    @classmethod
    @abstractmethod
    async def find_one(cls, *args, **kwargs) -> Self | None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def find(cls, *args, **kwargs) -> list[Self] | Cursor:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def first(cls, *args, **kwargs) -> Self | None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def last(cls, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def aggregate(cls, *args, **kwargs) -> Iterator[dict]:
        raise NotImplementedError

    # # # # # Count # # # # #
    @classmethod
    @abstractmethod
    async def count(cls, *args, **kwargs) -> int:
        raise NotImplementedError

    # # # # # Insert # # # # #
    @classmethod
    @abstractmethod
    async def insert_one(cls, *args, **kwargs) -> Self:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def insert_many(cls, *args, **kwargs) -> list[Self]:
        raise NotImplementedError

    # # # # # Delete # # # # #
    @abstractmethod
    async def delete(self) -> None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def delete_one(cls, *args, **kwargs) -> bool:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def delete_many(cls, *args, **kwargs) -> int:
        raise NotImplementedError

    # # # # # Update # # # # #
    @abstractmethod
    async def update(self, *args, **kwargs) -> None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def update_one(cls, *args, **kwargs) -> bool:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def update_many(cls, *args, **kwargs) -> int:
        raise NotImplementedError



from unittest import TestCase

import orjson as json

from panther import Panther, status
from panther.configs import config
from panther.test import WebsocketClient
from panther.websocket import GenericWebsocket


class WithoutAcceptWebsocket(GenericWebsocket):
    async def connect(self):
        pass


class CloseOnConnectWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.close()


class CustomCloseWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.close(code=status.WS_1013_TRY_AGAIN_LATER, reason='Come Back Later')


class MessageOnConnectWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.send('Hello')
        await self.close()


class QueryParamWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.send(self.query_params)
        await self.close()


class PathVariableWebsocket(GenericWebsocket):
    async def connect(self, name: str, age: int, is_male: bool):
        await self.accept()
        await self.send(
            f'{type(name).__name__}({name}), {type(age).__name__}({age}), {type(is_male).__name__}({is_male})'
        )
        await self.close()


class SendAllTypesWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        # Nothing
        await self.send()
        # None
        await self.send(None)
        # String
        await self.send('Hello Again')
        # Int
        await self.send(12)
        # Dict
        await self.send({'detail': 'ok'})
        # List
        await self.send([1, 2, 3, 'Ali', 4])
        # Tuple
        await self.send((1, 2, 3, 'Ali', 4))
        # Bytes
        await self.send(b'It Is Value Of A File')
        await self.close()


class WebsocketWithoutAuthentication(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.send(str(self.user))
        await self.close()


class WebsocketWithAuthentication(GenericWebsocket):
    auth = True

    async def connect(self):
        await self.accept()
        await self.send(self.user)
        await self.close()


class Permission:
    @classmethod
    async def authorization(cls, connection) -> bool:
        if connection.path == '/with-permission':
            return True
        else:
            return False


class WebsocketWithPermission(GenericWebsocket):
    permissions = [Permission]

    async def connect(self):
        await self.accept()
        await self.send('ok')
        await self.close()


class WebsocketWithoutPermission(GenericWebsocket):
    permissions = [Permission]

    async def connect(self):
        await self.accept()
        await self.send('no')
        await self.close()


urls = {
    'without-accept': WithoutAcceptWebsocket,
    'close-on-connect': CloseOnConnectWebsocket,
    'custom-close': CustomCloseWebsocket,
    'message-after-connect': MessageOnConnectWebsocket,
    'query-params': QueryParamWebsocket,
    'path-variable/<name>/<age>/<is_male>/': PathVariableWebsocket,
    'all-types': SendAllTypesWebsocket,
    'without-auth': WebsocketWithoutAuthentication,
    'with-auth': WebsocketWithAuthentication,
    'with-permission': WebsocketWithPermission,
    'without-permission': WebsocketWithoutPermission,
}


class TestWebsocket(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        config.HAS_WS = True
        cls.app = Panther(__name__, configs=__name__, urls=urls)

    def test_without_accept(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('without-accept')
        assert responses[0]['type'] == 'websocket.close'
        assert responses[0]['code'] == 1000
        assert responses[0]['reason'] == ''

    def test_close_on_connect(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('close-on-connect')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.close'
        assert responses[1]['code'] == 1000
        assert responses[1]['reason'] == ''

    def test_custom_close(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('custom-close')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.close'
        assert responses[1]['code'] == 1013
        assert responses[1]['reason'] == 'Come Back Later'

    def test_query_params(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('query-params', query_params={'name': 'ali', 'age': 27})
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        params = json.loads(responses[1]['text'])
        assert [*params.keys()] == ['name', 'age']
        assert params['name'] == 'ali'
        assert params['age'] == '27'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_message_after_connect(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('message-after-connect')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'Hello'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_path_variables(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('path-variable/Ali/25/true')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'str(Ali), int(25), bool(True)'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_all_types(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('all-types')

        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        # String
        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'Hello Again'
        assert 'bytes' not in responses[1]
        # Int
        assert responses[2]['type'] == 'websocket.send'
        assert responses[2]['text'] == '12'
        assert 'bytes' not in responses[2]
        # Dict
        assert responses[3]['type'] == 'websocket.send'
        assert responses[3]['text'] == '{"detail":"ok"}'
        assert 'bytes' not in responses[3]
        # List
        assert responses[4]['type'] == 'websocket.send'
        assert responses[4]['text'] == '[1,2,3,"Ali",4]'
        assert 'bytes' not in responses[4]
        # Tuple
        assert responses[5]['type'] == 'websocket.send'
        assert responses[5]['text'] == '[1,2,3,"Ali",4]'
        assert 'bytes' not in responses[5]
        # Bytes
        assert responses[6]['type'] == 'websocket.send'
        assert responses[6]['bytes'] == b'It Is Value Of A File'
        assert 'text' not in responses[6]

        # Close
        assert responses[7]['type'] == 'websocket.close'
        assert responses[7]['code'] == 1000
        assert responses[7]['reason'] == ''

    def test_without_auth(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('without-auth')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'None'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_with_auth_failed(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('with-auth')

        assert responses[0]['type'] == 'websocket.close'
        assert responses[0]['code'] == 1000
        assert responses[0]['reason'] == ''

    def test_with_auth_not_defined(self):
        ws = WebsocketClient(app=self.app)
        with self.assertLogs(level='CRITICAL') as captured:
            responses = ws.connect('with-auth?authorization=Bearer token')

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == '`WS_AUTHENTICATION` has not been set in configs'

        assert responses[0]['type'] == 'websocket.close'
        assert responses[0]['code'] == 1000
        assert responses[0]['reason'] == ''

    def test_with_auth_success(self):
        global WS_AUTHENTICATION, SECRET_KEY, DATABASE
        DATABASE = {
            'engine': {
                'class': 'panther.db.connections.PantherDBConnection',
            },
        }
        WS_AUTHENTICATION = 'panther.authentications.QueryParamJWTAuthentication'
        SECRET_KEY = 'hvdhRspoTPh1cJVBHcuingQeOKNc1uRhIP2k7suLe2g='
        token = 'eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyX2lkIjoxfQ.AF3nsj8IQ6t0ncqIx4quoyPfYaZ-pqUOW4z_euUztPM'
        app = Panther(__name__, configs=__name__, urls=urls)
        WS_AUTHENTICATION = None
        SECRET_KEY = None
        DATABASE = None

        ws = WebsocketClient(app=app)
        with self.assertLogs(level='ERROR') as captured:
            responses = ws.connect('with-auth', query_params={'authorization': f'Bearer {token}'})

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "User not found"'

        assert responses[0]['type'] == 'websocket.close'
        assert responses[0]['code'] == 1000
        assert responses[0]['reason'] == ''

    def test_with_permission(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('with-permission')

        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'ok'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_without_permission(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('without-permission')

        assert responses[0]['type'] == 'websocket.close'
        assert responses[0]['code'] == 1000
        assert responses[0]['reason'] == ''



from pathlib import Path
from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API
from panther.configs import config
from panther.db.models import BaseUser
from panther.request import Request
from panther.test import APIClient
from tests._utils import check_two_dicts


@API()
async def without_auth(request: Request):
    return request.user


@API(auth=True)
async def auth_required(request: Request):
    return request.user


urls = {
    'without': without_auth,
    'auth-required': auth_required,
}


class User(BaseUser):
    username: str
    password: str


AUTHENTICATION = 'panther.authentications.JWTAuthentication'
SECRET_KEY = 'hvdhRspoTPh1cJVBHcuingQeOKNc1uRhIP2k7suLe2g='
DB_PATH = 'test.pdb'
DATABASE = {
    'engine': {
        'class': 'panther.db.connections.PantherDBConnection',
        'path': DB_PATH,
    },
}
USER_MODEL = 'tests.test_authentication.User'


class TestAuthentication(IsolatedAsyncioTestCase):
    SHORT_TOKEN = {'Authorization': 'Token TOKEN'}
    NOT_ENOUGH_SEGMENT_TOKEN = {'Authorization': 'Bearer XXX'}
    JUST_BEARER_TOKEN = {'Authorization': 'Bearer'}
    BAD_UNICODE_TOKEN = {'Authorization': 'Bearer علی'}
    BAD_SIGNATURE_TOKEN = {'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJpZCI6MX0.JAWUkAU2mWhxcd6MS8r9pd44yBIfkEBmpr3WLeqIccM'}
    TOKEN_WITHOUT_USER_ID = {'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJpZCI6MX0.PpyXW0PgmGSPaaNirm_Ei4Y2fw9nb4TN26RN1u9RHSo'}
    TOKEN = {'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyX2lkIjoxfQ.AF3nsj8IQ6t0ncqIx4quoyPfYaZ-pqUOW4z_euUztPM'}

    def setUp(self) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        self.client = APIClient(app=app)

    def tearDown(self) -> None:
        Path(DB_PATH).unlink()

    async def test_user_without_auth(self):
        res = await self.client.get('without')
        assert res.status_code == 200
        assert res.data is None

        res = await self.client.get('without', headers={'Authorization': 'Token'})
        assert res.status_code == 200
        assert res.data is None

    async def test_user_auth_required_without_auth_class(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = None
        with self.assertLogs(level='CRITICAL') as captured:
            res = await self.client.get('auth-required')
        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == '"AUTHENTICATION" has not been set in configs'
        assert res.status_code == 500
        assert res.data['detail'] == 'Internal Server Error'
        config.AUTHENTICATION = auth_config

    async def test_user_auth_required_without_token(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required')

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Authorization is required"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_bad_token_1(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.SHORT_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Authorization keyword is not valid"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_bad_token2(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.JUST_BEARER_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Authorization should have 2 part"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_bad_token3(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.BAD_UNICODE_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == (
            'JWT Authentication Error: "\'latin-1\' codec can\'t encode characters in position 0-2: '
            'ordinal not in range(256)"'
        )
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_bad_token4(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.NOT_ENOUGH_SEGMENT_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Not enough segments"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_invalid_token_signature(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.BAD_SIGNATURE_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Signature verification failed."'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_token_without_user_id(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.TOKEN_WITHOUT_USER_ID)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Payload does not have `user_id`"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_token_user_not_found(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "User not found"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_token(self):
        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('auth-required', headers={'Authorization': f'Bearer {tokens["access_token"]}'})

        expected_response = {
            'username': 'Username',
            'password': 'Password'
        }
        assert res.status_code == 200
        res.data.pop('id')
        assert check_two_dicts(res.data, expected_response)



import contextlib
import os
from datetime import datetime
from typing import Annotated

from pydantic import Field, WrapValidator, PlainSerializer, BaseModel as PydanticBaseModel

from panther.configs import config
from panther.db.queries import Query
from panther.utils import scrypt, URANDOM_SIZE, timezone_now

with contextlib.suppress(ImportError):
    # Only required if user wants to use mongodb
    import bson


def validate_object_id(value, handler):
    if config.DATABASE.__class__.__name__ == 'MongoDBConnection':
        if isinstance(value, bson.ObjectId):
            return value
        else:
            try:
                return bson.ObjectId(value)
            except Exception as e:
                msg = 'Invalid ObjectId'
                raise ValueError(msg) from e
    return str(value)


ID = Annotated[str, WrapValidator(validate_object_id), PlainSerializer(lambda x: str(x), return_type=str)]


class Model(PydanticBaseModel, Query):
    def __init_subclass__(cls, **kwargs):
        if cls.__module__ == 'panther.db.models' and cls.__name__ == 'BaseUser':
            return
        config.MODELS.append(cls)

    id: ID | None = Field(None, validation_alias='_id')

    @property
    def _id(self):
        """
        return
            `str` for PantherDB
            `ObjectId` for MongoDB
        """
        return self.id


class BaseUser(Model):
    password: str = Field('', max_length=64)
    last_login: datetime | None = None
    date_created: datetime | None = Field(default_factory=timezone_now)

    async def update_last_login(self) -> None:
        await self.update(last_login=timezone_now())

    async def login(self) -> dict:
        """Return dict of access and refresh token"""
        return config.AUTHENTICATION.login(str(self.id))

    async def logout(self) -> dict:
        return await config.AUTHENTICATION.logout(self._auth_token)

    def set_password(self, password: str):
        """
        URANDOM_SIZE = 16 char -->
            salt = 16 bytes
            salt.hex() = 32 char
            derived_key = 32 char
        """
        salt = os.urandom(URANDOM_SIZE)
        derived_key = scrypt(password=password, salt=salt, digest=True)

        self.password = f'{salt.hex()}{derived_key}'

    def check_password(self, new_password: str) -> bool:
        size = URANDOM_SIZE * 2
        salt = self.password[:size]
        stored_hash = self.password[size:]
        derived_key = scrypt(password=new_password, salt=bytes.fromhex(salt), digest=True)

        return derived_key == stored_hash
